/*
Licensed to the Apache Software Foundation (ASF) under one or more
contributor license agreements.  See the NOTICE file distributed with
this work for additional information regarding copyright ownership.
The ASF licenses this file to You under the Apache License, Version 2.0
(the "License"); you may not use this file except in compliance with
the License.  You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package api

import (
	"encoding/json"
	"github.com/spf13/cast"
	"reflect"
	"time"

	"github.com/apache/incubator-devlake/core/errors"
	"github.com/apache/incubator-devlake/core/models"
	"github.com/apache/incubator-devlake/core/models/common"
	"github.com/go-playground/validator/v10"

	"github.com/mitchellh/mapstructure"
)

var defaultCustomDecoderHooks = []mapstructure.DecodeHookFunc{decodeHookStringFloat64, decodeHookStringToTime, DecodeHook}

func decodeHookStringFloat64(f reflect.Type, t reflect.Type, data interface{}) (interface{}, error) {
	if f.Kind() == reflect.Float64 && t.Kind() == reflect.Struct && t == reflect.TypeOf(common.StringFloat64{}) {
		return *common.NewStringFloat64FromAny(data), nil
	}
	if f.Kind() == reflect.Float64 && t.Kind() == reflect.Ptr && t == reflect.TypeOf(&common.StringFloat64{}) {
		return common.NewStringFloat64FromAny(data), nil
	}
	return data, nil
}

func decodeHookStringToTime(f reflect.Type, t reflect.Type, data interface{}) (interface{}, error) {
	if f.Kind() == reflect.String {
		if t.Kind() == reflect.Struct && t == reflect.TypeOf(time.Time{}) {
			timeData, err := common.ConvertStringToTime(cast.ToString(data))
			if err != nil {
				return data, err
			}
			return timeData, nil
		}
		if t.Kind() == reflect.Ptr && t == reflect.TypeOf(&time.Time{}) {
			timeData, err := common.ConvertStringToTime(cast.ToString(data))
			if err != nil {
				return data, err
			}
			return &timeData, nil
		}
	}
	return data, nil
}

func DecodeHook(f reflect.Type, t reflect.Type, data interface{}) (interface{}, error) {
	if data == nil {
		return nil, nil
	}
	if t == reflect.TypeOf(json.RawMessage{}) {
		return json.Marshal(data)
	}
	if t != reflect.TypeOf(common.Iso8601Time{}) && t != reflect.TypeOf(time.Time{}) {
		return data, nil
	}

	var tt time.Time
	var err error
	switch f.Kind() {
	case reflect.String:
		tt, err = common.ConvertStringToTime(data.(string))
	case reflect.Float64:
		tt = time.Unix(0, int64(data.(float64))*int64(time.Millisecond))
	case reflect.Int64:
		tt = time.Unix(0, data.(int64)*int64(time.Millisecond))
	case reflect.Struct:
		if f == reflect.TypeOf(time.Time{}) || f == reflect.TypeOf(common.Iso8601Time{}) {
			return data, nil
		}
	case reflect.Ptr:
		if f == reflect.TypeOf(&time.Time{}) || f == reflect.TypeOf(&common.Iso8601Time{}) {
			return data, nil
		}
	}
	if err != nil {
		return data, nil
	}
	switch t {
	case reflect.TypeOf(common.Iso8601Time{}):
		return common.Iso8601Time{Time: tt}, nil
	case reflect.TypeOf(&common.Iso8601Time{}):
		return &common.Iso8601Time{Time: tt}, nil
	}
	return data, nil
}

// DecodeMapStruct with time.Time and Iso8601Time support
func DecodeMapStruct(input map[string]interface{}, result interface{}, zeroFields bool) errors.Error {
	result = models.UnwrapObject(result)
	decoder, err := mapstructure.NewDecoder(&mapstructure.DecoderConfig{
		ZeroFields:       zeroFields,
		DecodeHook:       mapstructure.ComposeDecodeHookFunc(defaultCustomDecoderHooks...),
		Result:           result,
		WeaklyTypedInput: true,
	})
	if err != nil {
		return errors.Convert(err)
	}
	if err := decoder.Decode(input); err != nil {
		return errors.Convert(err)
	}
	return errors.Convert(err)
}

// Decode decodes `source` into `target`. Pass an optional validator to validate the target.
func Decode(source interface{}, target interface{}, vld *validator.Validate) errors.Error {
	target = models.UnwrapObject(target)
	if err := mapstructure.Decode(source, &target); err != nil {
		return errors.Default.Wrap(err, "error decoding map into target type")
	}
	if vld != nil {
		if err := vld.Struct(target); err != nil {
			return errors.Default.Wrap(err, "error validating target")
		}
	}
	return nil
}
