/*
Licensed to the Apache Software Foundation (ASF) under one or more
contributor license agreements.  See the NOTICE file distributed with
this work for additional information regarding copyright ownership.
The ASF licenses this file to You under the Apache License, Version 2.0
(the "License"); you may not use this file except in compliance with
the License.  You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package utils

import (
	"encoding/json"
	"reflect"
	"time"

	"github.com/apache/incubator-devlake/core/models"
	"github.com/apache/incubator-devlake/core/models/common"

	"github.com/apache/incubator-devlake/core/errors"
	"github.com/go-playground/validator/v10"

	"github.com/mitchellh/mapstructure"
)

func DecodeHook(f reflect.Type, t reflect.Type, data interface{}) (interface{}, error) {
	if data == nil {
		return nil, nil
	}
	if t == reflect.TypeOf(json.RawMessage{}) {
		return json.Marshal(data)
	}
	// to support decoding url.Values (query string) to non-array variables
	if t.Kind() != reflect.Slice && t.Kind() != reflect.Array &&
		(f.Kind() == reflect.Slice || f.Kind() == reflect.Array) {
		v := reflect.ValueOf(data)
		if v.Len() == 1 {
			data = v.Index(0).Interface()
			var result interface{}
			err := DecodeMapStruct(data, &result, true)
			return result, err
		}
	}

	if t != reflect.TypeOf(common.Iso8601Time{}) && t != reflect.TypeOf(time.Time{}) {
		return data, nil
	}

	var tt time.Time
	var err error

	switch f.Kind() {
	case reflect.String:
		tt, err = common.ConvertStringToTime(data.(string))
	case reflect.Float64:
		tt = time.Unix(0, int64(data.(float64))*int64(time.Millisecond))
	case reflect.Int64:
		tt = time.Unix(0, data.(int64)*int64(time.Millisecond))
	}
	if err != nil {
		return data, nil
	}

	if t == reflect.TypeOf(common.Iso8601Time{}) {
		return common.Iso8601Time{Time: tt}, nil
	}
	return tt, nil
}

// DecodeMapStruct with time.Time and Iso8601Time support
func DecodeMapStruct(input interface{}, result interface{}, zeroFields bool) errors.Error {
	result = models.UnwrapObject(result)
	decoder, err := mapstructure.NewDecoder(&mapstructure.DecoderConfig{
		ZeroFields:       zeroFields,
		DecodeHook:       mapstructure.ComposeDecodeHookFunc(DecodeHook),
		Result:           result,
		WeaklyTypedInput: true,
	})
	if err != nil {
		return errors.Convert(err)
	}

	if err := decoder.Decode(input); err != nil {
		return errors.Convert(err)
	}
	return errors.Convert(err)
}

// Decode decodes `source` into `target`. Pass an optional validator to validate the target.
func Decode(source interface{}, target interface{}, vld *validator.Validate) errors.Error {
	target = models.UnwrapObject(target)
	if err := mapstructure.Decode(source, &target); err != nil {
		return errors.Default.Wrap(err, "error decoding map into target type")
	}
	if vld != nil {
		if err := vld.Struct(target); err != nil {
			return errors.Default.Wrap(err, "error validating target")
		}
	}
	return nil
}
